/* 
 * Copyright (c) 2012, Fraunhofer-Gesellschaft
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * (1) Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the disclaimer at the end.
 *     Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 
 * (2) Neither the name of Fraunhofer nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */
package org.ogf.graap.wsag.server.rest;

import java.net.URI;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.List;
import java.util.Vector;

import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.core.UriBuilder;
import javax.ws.rs.core.UriInfo;

import org.apache.log4j.Logger;
import org.apache.velocity.exception.ResourceNotFoundException;
import org.apache.xmlbeans.QNameSet;
import org.apache.xmlbeans.XmlObject;
import org.ogf.graap.wsag.api.Agreement;
import org.ogf.graap.wsag.api.Negotiation;
import org.ogf.graap.wsag.api.PendingAgreementListener;
import org.ogf.graap.wsag.api.exceptions.AgreementFactoryException;
import org.ogf.graap.wsag.api.exceptions.NegotiationFactoryException;
import org.ogf.graap.wsag.api.exceptions.ResourceUnavailableException;
import org.ogf.graap.wsag.api.exceptions.ResourceUnknownException;
import org.ogf.graap.wsag.api.rest.RestAgreement;
import org.ogf.graap.wsag.api.rest.RestAgreementFactory;
import org.ogf.graap.wsag.api.rest.RestNegotiation;
import org.ogf.graap.wsag.api.types.AgreementOfferType;
import org.ogf.graap.wsag.client.api.AgreementAcceptanceClient;
import org.ogf.graap.wsag.server.api.AgreementAcceptanceFactory;
import org.ogf.graap.wsag.server.persistence.PersistentAgreement;
import org.ogf.graap.wsag.server.persistence.PersistentAgreementFactory;
import org.ogf.graap.wsag.server.rest.acceptance.RestAcceptanceRegistryFacade;
import org.ogf.graap.wsag.server.rest.acceptance.RestAgreementAcceptanceFactory;
import org.ogf.schemas.graap.wsAgreement.AgreementFactoryPropertiesDocument;
import org.ogf.schemas.graap.wsAgreement.AgreementTemplateType;
import org.ogf.schemas.graap.wsAgreement.AgreementType;
import org.ogf.schemas.graap.wsAgreement.CreateAgreementInputDocument;
import org.ogf.schemas.graap.wsAgreement.CreatePendingAgreementInputDocument;
import org.ogf.schemas.graap.wsAgreement.NoncriticalExtensionType;
import org.ogf.schemas.graap.wsAgreement.negotiation.InitiateNegotiationInputDocument;
import org.ogf.schemas.graap.wsAgreement.negotiation.InitiateNegotiationOutputDocument;
import org.ogf.schemas.graap.wsAgreement.negotiation.NegotiationContextType;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

/**
 * @author owaeld
 * 
 */
public class RestAgreementFactoryFacade
    implements RestAgreementFactory
{
    private static final Logger LOG = Logger.getLogger( RestAgreementFactoryFacade.class );

    private final String factoryId;

    private final PersistentAgreementFactory factory;

    private final UriInfo uriInfo;

    private UriBuilder uriBuilder;

    public RestAgreementFactoryFacade( String factoryId, PersistentAgreementFactory factory, UriInfo uriInfo )
    {
        this.factoryId = factoryId;
        this.factory = factory;
        this.uriInfo = uriInfo;
    }

    private PersistentAgreementFactory getFactory()
        throws ResourceUnknownException, ResourceUnavailableException
    {
        return factory;
    }

    @Path( "/Acceptance" )
    public RestAcceptanceRegistryFacade getAgreementAcceptance()
    {
        return new RestAcceptanceRegistryFacade();
    }

    @Path( "/Agreements/{agreementId}" )
    public RestAgreement getAgreement( @PathParam( "agreementId" ) String agreementId )
        throws ResourceUnknownException, ResourceUnavailableException
    {
        try
        {
            //
            // lookup the factory with the given id from the WSAG4J engine
            //
            PersistentAgreement agreementInstance = getFactory().find( agreementId );

            //
            // if the engine returns null, we raise a resource not found exception
            //
            if ( agreementInstance == null )
            {
                String message =
                    MessageFormat.format( "agreement resource with id ''{0}'' is unknown.", agreementId );

                throw new ResourceUnknownException( message );
            }
            return new RestAgreementFacade( agreementInstance, getFactory() );
        }

        catch ( ResourceUnknownException e )
        {
            throw e;
        }
        catch ( Exception e )
        {
            throw new ResourceUnavailableException( "failed to lookup agreement resource", e );
        }
    }

    /**
     * Returns the requested negotiation resource.
     * 
     * @param negotiationId
     *            the negotiation id
     * 
     * @return the resolved negotiation resource
     */
    @Path( "/Negotiations/{negotiationId}" )
    public RestNegotiation getNegotiation( @PathParam( "negotiationId" ) String negotiationId )
    {
        Negotiation negotiation = RestNegotiationRegistry.get( negotiationId );
        if ( negotiation == null )
        {
            String message = MessageFormat.format( "negotiation resource ''{0}'' not found", negotiationId );
            throw new ResourceNotFoundException( message );
        }

        return new RestNegotiationFacade( factory.getResourceId(), negotiationId );
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ogf.graap.wsag.api.rest.RestAgreementFactory#getResourceProperties()
     */
    @Override
    public AgreementFactoryPropertiesDocument getResourceProperties()
        throws ResourceUnknownException, ResourceUnavailableException
    {
        AgreementFactoryPropertiesDocument properties =
            AgreementFactoryPropertiesDocument.Factory.newInstance();
        AgreementTemplateType[] templates = getFactory().getTemplates();
        properties.addNewAgreementFactoryProperties().setTemplateArray( templates );
        return properties;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ogf.graap.wsag.api.rest.RestAgreementFactory#getTemplates()
     */
    @Override
    public AgreementFactoryPropertiesDocument getTemplates()
        throws ResourceUnknownException, ResourceUnavailableException
    {
        AgreementTemplateType[] templates = getFactory().getTemplates();
        AgreementFactoryPropertiesDocument properties =
            AgreementFactoryPropertiesDocument.Factory.newInstance();
        properties.addNewAgreementFactoryProperties().setTemplateArray( templates );
        return properties;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ogf.graap.wsag.api.rest.RestAgreementFactory#listAgreements()
     */
    @Override
    public List<URI> listAgreements() throws ResourceUnknownException, ResourceUnavailableException
    {
        List<URI> uris = new Vector<URI>();
        PersistentAgreementFactory factory = getFactory();

        try
        {
            PersistentAgreement[] agreements = factory.list();
            for ( int i = 0; i < agreements.length; i++ )
            {
                String agreementId = agreements[i].getAgreement().getAgreementId();

                URI agreementUri = getUriBuilder().clone().path( agreementId ).build();

                uris.add( agreementUri );
            }
        }
        catch ( Exception e )
        {
            throw new ResourceUnavailableException( e );
        }

        return uris;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ogf.graap.wsag.api.rest.RestAgreementFactory#getResourceId()
     */
    @Override
    public String getResourceId() throws ResourceUnknownException, ResourceUnavailableException
    {
        return getFactory().getResourceId();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ogf.graap.wsag.api.rest.RestAgreementFactory#createAgreement(org.ogf.schemas.graap.wsAgreement.
     * CreateAgreementInputDocument)
     */
    @Override
    public URI createAgreement( CreateAgreementInputDocument input )
        throws AgreementFactoryException, ResourceUnknownException, ResourceUnavailableException
    {
        AgreementType agreementOffer = input.getCreateAgreementInput().getAgreementOffer();

        HashMap<String, Object> context = new HashMap<String, Object>();

        try
        {
            RestAgreementAcceptanceFactory acceptanceFactory =
                RestAgreementAcceptanceFactory.newInstance( getAcceptanceUriBuilder() );
            context.put( AgreementAcceptanceFactory.AGREEMENT_ACCEPTANCE_FACTORY, acceptanceFactory );
        }
        catch ( Exception e )
        {
            LOG.warn( "failed to add acceptance factory to invocation context", e );
        }

        Agreement agreement =
            getFactory().createAgreement( new AgreementOfferType( agreementOffer ), context );

        String agreementId = agreement.getAgreementId();
        URI agreementUri = getUriBuilder().clone().path( agreementId ).build();

        return agreementUri;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ogf.graap.wsag.api.rest.RestAgreementFactory#createPendingAgreement(org.ogf.schemas.graap.wsAgreement
     * .CreatePendingAgreementInputDocument)
     */
    @Override
    public URI createPendingAgreement( CreatePendingAgreementInputDocument input )
        throws AgreementFactoryException, ResourceUnknownException, ResourceUnavailableException
    {
        AgreementType offer = input.getCreatePendingAgreementInput().getAgreementOffer();

        PendingAgreementListener listener = null;
        HashMap<String, Object> context = new HashMap<String, Object>();

        try
        {
            RestAgreementAcceptanceFactory acceptanceFactory =
                RestAgreementAcceptanceFactory.newInstance( getAcceptanceUriBuilder() );
            context.put( AgreementAcceptanceFactory.AGREEMENT_ACCEPTANCE_FACTORY, acceptanceFactory );
        }
        catch ( Exception e )
        {
            LOG.warn( "failed to add acceptance factory to invocation context", e );
        }

        EndpointReferenceType acceptanceEPR =
            input.getCreatePendingAgreementInput().getAgreementAcceptanceEPR();

        if ( acceptanceEPR != null )
        {

            try
            {
                final AgreementAcceptanceClient client =
                    AgreementAcceptanceClient.FACTORY.newInstance( acceptanceEPR, null );

                listener = new PendingAgreementListener()
                {

                    @Override
                    public void reject() throws Exception
                    {
                        client.reject();
                    }

                    @Override
                    public void accept() throws Exception
                    {
                        client.accept();
                    }
                };
            }
            catch ( Exception e )
            {
                LOG.error( "failed to register acceptance client", e );
            }
        }

        Agreement agreement =
            getFactory().createPendingAgreement( new AgreementOfferType( offer ), listener, context );

        String agreementId = agreement.getAgreementId();
        URI agreementUri = getUriBuilder().clone().path( agreementId ).build();

        return agreementUri;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ogf.graap.wsag.api.rest.RestAgreementFactory#initiateNegotiation(org.ogf.schemas.graap.wsAgreement
     * .negotiation.InitiateNegotiationInputDocument)
     */
    @Override
    public URI initiateNegotiation( InitiateNegotiationInputDocument input )
        throws NegotiationFactoryException, ResourceUnknownException, ResourceUnavailableException
    {
        NegotiationContextType context = input.getInitiateNegotiationInput().getNegotiationContext();
        NoncriticalExtensionType[] noncriticalExtensions =
            input.getInitiateNegotiationInput().getNoncriticalExtensionArray();

        String namespaceURI =
            InitiateNegotiationOutputDocument.type.getDocumentElementName().getNamespaceURI();
        QNameSet qNameSet = QNameSet.forWildcardNamespaceString( "##other", namespaceURI );
        XmlObject[] criticalExtensions = input.getInitiateNegotiationInput().selectChildren( qNameSet );

        Negotiation negotiation =
            getFactory().initiateNegotiation( context, criticalExtensions, noncriticalExtensions,
                new HashMap<String, Object>() );

        String uuid = RestNegotiationRegistry.add( negotiation );

        URI negotiationUri = getUriBuilder().clone().path( uuid ).build();

        return negotiationUri;
    }

    private UriBuilder getAcceptanceUriBuilder()
    {
        return uriInfo.getBaseUriBuilder().path( factoryId ).path( "Acceptance" );
    }

    private UriBuilder getUriBuilder()
    {
        return uriInfo.getAbsolutePathBuilder();
    }

}
