/* 
 * Copyright (c) 2012, Fraunhofer-Gesellschaft
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * (1) Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the disclaimer at the end.
 *     Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 
 * (2) Neither the name of Fraunhofer nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */
package org.ogf.graap.wsag.server.rest;

import javax.ws.rs.Path;
import javax.ws.rs.PathParam;

import org.apache.xmlbeans.QNameSet;
import org.apache.xmlbeans.XmlObject;
import org.ogf.graap.wsag.api.Negotiation;
import org.ogf.graap.wsag.api.exceptions.NegotiationException;
import org.ogf.graap.wsag.api.exceptions.ResourceUnavailableException;
import org.ogf.graap.wsag.api.exceptions.ResourceUnknownException;
import org.ogf.graap.wsag.api.rest.RestNegotiation;
import org.ogf.schemas.graap.wsAgreement.AgreementTemplateType;
import org.ogf.schemas.graap.wsAgreement.negotiation.AdvertiseInputDocument;
import org.ogf.schemas.graap.wsAgreement.negotiation.NegotiateInputDocument;
import org.ogf.schemas.graap.wsAgreement.negotiation.NegotiateOutputDocument;
import org.ogf.schemas.graap.wsAgreement.negotiation.NegotiationContextType;
import org.ogf.schemas.graap.wsAgreement.negotiation.NegotiationOfferType;
import org.ogf.schemas.graap.wsAgreement.negotiation.NegotiationPropertiesDocument;
import org.ogf.schemas.graap.wsAgreement.negotiation.TerminateInputDocument;

/**
 * @author owaeld
 * 
 */
@Path( "/factories/{factoryId}/Negotiations/{negotiationId}" )
public class RestNegotiationFacade
    implements RestNegotiation
{
    private final String factoryId;

    private final String negotiationId;

    public RestNegotiationFacade( @PathParam( "factoryId" ) String factoryId,
                                  @PathParam( "negotiationId" ) String negotiationId )
    {
        this.factoryId = factoryId;
        this.negotiationId = negotiationId;
    }

    private Negotiation getNegotiation()
    {
        /*
         * each negotiation has a global unique id, no need to use the factory id here
         */
        return RestNegotiationRegistry.get( negotiationId );
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ogf.graap.wsag.api.rest.RestNegotiation#getNegotiationContext()
     */
    @Override
    public NegotiationPropertiesDocument getNegotiationContext()
        throws ResourceUnknownException, ResourceUnavailableException
    {
        NegotiationContextType context = getNegotiation().getNegotiationContext();

        NegotiationPropertiesDocument properties = NegotiationPropertiesDocument.Factory.newInstance();
        properties.addNewNegotiationProperties().addNewNegotiationContext().set( context );

        return properties;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ogf.graap.wsag.api.rest.RestNegotiation#getNegotiableTemplates()
     */
    @Override
    public NegotiationPropertiesDocument getNegotiableTemplates()
        throws ResourceUnknownException, ResourceUnavailableException
    {
        AgreementTemplateType[] templates = getNegotiation().getNegotiableTemplates();

        NegotiationPropertiesDocument properties = NegotiationPropertiesDocument.Factory.newInstance();
        properties.addNewNegotiationProperties().setNegotiableTemplateArray( templates );

        return properties;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ogf.graap.wsag.api.rest.RestNegotiation#getNegotiationOffers()
     */
    @Override
    public NegotiationPropertiesDocument getNegotiationOffers()
        throws ResourceUnknownException, ResourceUnavailableException
    {
        NegotiationOfferType[] offers = getNegotiation().getNegotiationOffers();

        NegotiationPropertiesDocument properties = NegotiationPropertiesDocument.Factory.newInstance();
        properties.addNewNegotiationProperties().setNegotiationOfferArray( offers );

        return properties;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ogf.graap.wsag.api.rest.RestNegotiation#negotiate(org.ogf.schemas.graap.wsAgreement.negotiation
     * .NegotiateInputDocument)
     */
    @Override
    public NegotiateOutputDocument negotiate( NegotiateInputDocument input )
        throws NegotiationException, ResourceUnknownException, ResourceUnavailableException
    {
        NegotiationOfferType[] offers = input.getNegotiateInput().getNegotiationOfferArray();
        QNameSet others =
            QNameSet.forWildcardNamespaceString( "##other",
                NegotiateInputDocument.type.getDocumentElementName().getNamespaceURI() );
        XmlObject[] noncritical = input.getNegotiateInput().selectChildren( others );

        NegotiationOfferType[] counterOffers = getNegotiation().negotiate( offers, noncritical );

        NegotiateOutputDocument output = NegotiateOutputDocument.Factory.newInstance();
        output.addNewNegotiateOutput().setNegotiationCounterOfferArray( counterOffers );

        return output;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ogf.graap.wsag.api.rest.RestNegotiation#advertise(org.ogf.schemas.graap.wsAgreement.negotiation
     * .AdvertiseInputDocument)
     */
    @Override
    public void advertise( AdvertiseInputDocument input )
        throws NegotiationException, ResourceUnknownException, ResourceUnavailableException
    {
        NegotiationOfferType[] offers = input.getAdvertiseInput().getNegotiationOfferArray();
        QNameSet others =
            QNameSet.forWildcardNamespaceString( "##other",
                NegotiateInputDocument.type.getDocumentElementName().getNamespaceURI() );
        XmlObject[] noncritical = input.getAdvertiseInput().selectChildren( others );

        getNegotiation().advertise( offers, noncritical );
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ogf.graap.wsag.api.rest.RestNegotiation#terminate(org.ogf.schemas.graap.wsAgreement.negotiation
     * .TerminateInputDocument)
     */
    @Override
    public void terminate( TerminateInputDocument reason )
        throws ResourceUnknownException, ResourceUnavailableException
    {
        getNegotiation().terminate();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ogf.graap.wsag.api.rest.RestNegotiation#destroy()
     */
    @Override
    public void destroy() throws ResourceUnknownException, ResourceUnavailableException
    {
        RestNegotiationRegistry.remove( negotiationId );
    }

}
