/* 
 * Copyright (c) 2012, Fraunhofer-Gesellschaft
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * (1) Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the disclaimer at the end.
 *     Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 
 * (2) Neither the name of Fraunhofer nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */
package org.ogf.graap.wsag.server.rest;

import java.beans.Introspector;
import java.sql.Driver;
import java.sql.DriverManager;
import java.util.Enumeration;

import javax.servlet.ServletContext;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;

import org.apache.log4j.Logger;
import org.apache.xml.resolver.tools.CatalogResolver;
import org.hsqldb.DatabaseManager;
import org.ogf.graap.wsag.server.persistence.EmfRegistry;

/**
 * @author owaeld
 * 
 */
public class WSAG4JContextListener
    implements ServletContextListener
{

    /**
     * 
     */
    private static final int DB_WAITING_TIME = 2500;

    /**
     * 
     */
    private static final int MAX_DB_SHUTDOWN_RETRIES = 25;

    private static final Logger LOG = Logger.getLogger( WSAG4JContextListener.class );

    /**
     * {@inheritDoc}
     * 
     * @see javax.servlet.ServletContextListener#contextDestroyed(javax.servlet.ServletContextEvent)
     */
    @Override
    public void contextDestroyed( ServletContextEvent eContext )
    {

        // shutdown the WSAG4J instance (the server)
        try
        {
            RestWsagEngine.shutdown();
        }
        catch ( Exception e )
        {
            LOG.error( "Failed to cleanly shutdown WSAG4J engine. Proceed with shutdown process.", e );
        }

        //
        // in case there is a WSAG4J version without hsqldb, remove this block
        //
        //
        // hard shutdown of the HSQL Timer on undeploy/stop of the webapp
        //
        Thread runner = DatabaseManager.getTimer().getThread();
        DatabaseManager.getTimer().shutdown();

        //
        // wait for HSQLDB to finish outstanding work.
        //
        // HSQLDB has a default write delay of 60 seconds for scripts.
        // Therefore we wait a bit more than a minute at max for HSQLDB
        // shutdown.
        //
        int maxWaitingTime = MAX_DB_SHUTDOWN_RETRIES;

        while ( runner.isAlive() )
        {
            if ( maxWaitingTime == 0 )
            {
                break;
            }

            maxWaitingTime--;

            try
            {
                LOG.debug( "wait for HSQLDB to finish outstanding work..." );
                synchronized ( this )
                {
                    wait( DB_WAITING_TIME );
                }
            }
            catch ( InterruptedException e )
            {
                break;
            }
        }

        //
        // check if HSQLDB terminated normally
        //
        if ( runner.isAlive() )
        {
            LOG.warn( "HSQLDB failed to terminate normally. Send interupt." );
            runner.interrupt();
        }
        else
        {
            LOG.info( "HSQLDB shutdown complete." );
        }
        //
        // end of hsqldb specific block
        //

        LOG.debug( "cleanup database drivers..." );
        Introspector.flushCaches();
        for ( Enumeration<Driver> e = DriverManager.getDrivers(); e.hasMoreElements(); )
        {
            Driver driver = e.nextElement();
            if ( driver.getClass().getClassLoader() == getClass().getClassLoader() )
            {
                try
                {
                    LOG.debug( "deregister: " + driver.getClass().getName() );
                    DriverManager.deregisterDriver( driver );
                }
                catch ( Exception ex )
                {
                    LOG.warn( "failed to deregister driver.", ex );
                }
            }
        }

        LOG.info( "shutdown of WSAG4J service completed" );
    }

    /**
     * {@inheritDoc}
     * 
     * @see javax.servlet.ServletContextListener#contextInitialized(javax.servlet.ServletContextEvent)
     */
    @Override
    public void contextInitialized( ServletContextEvent contextEvent )
    {
        synchronized ( WSAG4JContextListener.class )
        {

            //
            // use the CatalogResolver as default for XmlBeans
            // this is required for template validation
            //
            System.setProperty( "xmlbean.entityResolver", CatalogResolver.class.getName() );

            try
            {
                //
                // get the real path for bootstraping configuration files, etc.
                //
                ServletContext context = contextEvent.getServletContext();
                String path = context.getRealPath( "/" );
                String datapath = path + "/WEB-INF/wsag4j-data";

                if ( !System.getProperties().containsKey( EmfRegistry.WSAG4J_DATAPATH ) )
                {
                    System.getProperties().setProperty( EmfRegistry.WSAG4J_DATAPATH, datapath );
                }

                //
                // create and initialize the isolation layer
                //
                initializeEngine();
            }
            catch ( Exception e )
            {
                String message = "WSAG4J initialization process failed. Error: " + e.getMessage();
                LOG.error( message );
            }

        }
    }

    private void initializeEngine()
    {
        try
        {
            LOG.debug( "Initialize REST wsag engine..." );

            //
            // initialize the WSAG4J server configuration
            //
            RestWsagEngine.startup();
        }
        catch ( Exception ex )
        {
            throw new RuntimeException( "The initialization process of the WSAG4J engine failed. Reason: "
                + ex.getMessage(), ex );
        }
    }
}
